#ifndef MEMORY_H_INCLUDED
#define MEMORY_H_INCLUDED

#include "pc6001v.h"
#include "breakpoint.h"
#include "device.h"
#include "ini.h"


#define MAIN_ROM_0	(MainRom)
#define MAIN_ROM_1	(MainRom+0x2000)
#define MAIN_ROM_2	(MainRom+0x4000)
#define MAIN_ROM_3	(MainRom+0x6000)
#define EXT_ROM_0	(ExtRom ? ExtRom        : EmptyMEM)
#define EXT_ROM_1	(ExtRom ? ExtRom+0x2000 : EmptyMEM)
#define KANJI_ROM_0	(KanjiRom)
#define KANJI_ROM_1	(KanjiRom+0x2000)
#define KANJI_ROM_2	(KanjiRom+0x4000)
#define KANJI_ROM_3	(KanjiRom+0x6000)
#define VOICE_ROM_0	(VoiceRom)
#define VOICE_ROM_1	(VoiceRom+0x2000)
#define MAIN_RAM_0	(MainRam)
#define MAIN_RAM_1	(MainRam+0x2000)
#define MAIN_RAM_2	(MainRam+0x4000)
#define MAIN_RAM_3	(MainRam+0x6000)
#define MAIN_RAM_4	(MainRam+0x8000)
#define MAIN_RAM_5	(MainRam+0xa000)
#define MAIN_RAM_6	(MainRam+0xc000)
#define MAIN_RAM_7	(MainRam+0xe000)
#define EXT_RAM_0	(EmptyMEM)
#define EXT_RAM_1	(EmptyMEM)
#define EXT_RAM_2	(EmptyMEM)
#define EXT_RAM_3	(EmptyMEM)
#define EXT_RAM_4	(EmptyMEM)
#define EXT_RAM_5	(EmptyMEM)
#define EXT_RAM_6	(EmptyMEM)
#define EXT_RAM_7	(EmptyMEM)
#define KJ_ROM_0	(kj_rom ? ( kj_LR ? KANJI_ROM_2 : KANJI_ROM_0 ) : VOICE_ROM_0 )
#define KJ_ROM_1	(KJ_ROM_0+0x2000)


// 
typedef struct{
	char *FileName;	// ե̾
	Uint32 Size;	// 
	Uint32 Crc;		// CRC32
	BYTE Init;		// ǡ
} MEMINFO;


////////////////////////////////////////////////////////////////
// 饹
////////////////////////////////////////////////////////////////
class MemBlock {
protected:
	BYTE MomoryBlock[0x2000];	// 8KBΥ֥å
	int Wait;					// 
	
public:
	MemBlock( BYTE );			// 󥹥ȥ饯
	virtual ~MemBlock();		// ǥȥ饯
	
	BYTE Read( WORD addr ){ return MomoryBlock[addr&0x1fff]; }	// ꡼
	virtual void Write( WORD, BYTE ){};							// 饤
	
	void SetWait( int wait ){ Wait = wait; }					// 
};

class RomBlock : public MemBlock {
protected:
	
public:
	RomBlock();		// 󥹥ȥ饯
	~RomBlock();	// ǥȥ饯
	
	BOOL LoadRom( char *, BYTE *, DWORD );		// ROMեɹߥ
};

class RamBlock : public MemBlock {
protected:
	
public:
	RamBlock();		// 󥹥ȥ饯
	~RamBlock();	// ǥȥ饯
	
	void Write( WORD addr , BYTE data ){ MomoryBlock[addr&0x1fff] = data; }	// 饤
};





class cP6memory : public IDoko {
protected:
	// 
	static const MEMINFO IEXTROM;
	static const MEMINFO IEMPTY;
	static const MEMINFO IDUMMY;
	
	BOOL CGBank;		// CG ROM BANK TRUE:ͭ FALSE:̵
	
	BYTE *MainRom;		// BASIC ROM	60:16KB 62,66:32KB
	BYTE *ExtRom;		// ĥ ROM		60:16KB 62,66:16KB
	BYTE *CGRom1;		// CG ROM1		60: 4KB 62,66: 8KB
	BYTE *MainRam;		// RAM			60:32KB 62,66:64KB
	BYTE *EmptyMEM;		// ̤ 60: 8KB 62,66: 8KB
	BYTE *DummyRAM;		// ߡRAM	60: 8KB 62,66: 8KB
	
	BYTE *rm_blk[8];	// ꡼ɻ֥åݥ(8KB*8)
	BYTE *wm_blk[8];	// 饤Ȼ֥åݥ(8KB*8)
	BYTE ram_sel[4];	// RAM,RAM (16KB*4)
	
	BOOL AllocMemory( BYTE **, const MEMINFO *, char * );	// ݤROMեɹ
	
	BOOL InitCommon( char *, BOOL );				// ʬ
	
	BOOL EnableChkCRC;								// CRCåͭ?
	static const Uint32 Crc32Table[];				// CRC32׻ѥơ֥
	static Uint32 CalcCrc32( Uint8 *, int );		// CRC32׻
	
	// 8255ϴϢؿ
	virtual void SetCGBank( BOOL );					// CG ROM BANK ڤؤ
	
public:
	cP6memory();									// 󥹥ȥ饯
	virtual ~cP6memory();							// ǥȥ饯
	
	virtual BOOL Init( char *, char *, BOOL ) = 0;	// 
	virtual void Reset(){};							// ꥻå
	
	BYTE Read( WORD );								// ꡼
	void Write( WORD, BYTE );						// 饤
	
	BOOL MountExtRom( char * );						// ĥROM ޥ
	void UnmountExtRom();							// ĥROM ޥ
	
	// ľܥؿ
	virtual BYTE ReadMainRom( WORD addr ){ return MainRom[addr&0x3fff]; }
	virtual BYTE ReadMainRam( WORD addr ){ return MainRam[addr&0x3fff]; }
	virtual BYTE ReadExtRom ( WORD addr ){ return ExtRom ? ExtRom[addr&0x3fff] : 0xff; }
	virtual BYTE ReadCGrom1 ( WORD addr ){ return CGRom1[addr&0x1fff]; }
	virtual BYTE ReadCGrom2 ( WORD )     { return 0xff; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ǥХåѴؿ
	virtual char *GetReadMemBlk( int, char * )  = 0;
	virtual char *GetWriteMemBlk( int, char * ) = 0;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
};


class cP6mem60 : public cP6memory, public Device {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO IMAINRAM;
	
	BOOL useextram;					// ĥRAMȤ?
	
	void SetMB_60();				// ֥å(60Τ)
	
	// ǥХ
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/Oؿ
	inline void Out92H( DWORD, BYTE data ){ SetCGBank( data&4 ? FALSE : TRUE ); }
	inline void Out93H( DWORD, BYTE data ){ if( (data&0x8e) == 4 ) SetCGBank( data&1 ? FALSE : TRUE ); }
	
public:
	cP6mem60( const ID&, BOOL );				// 󥹥ȥ饯
	~cP6mem60();								// ǥȥ饯
	
	virtual BOOL Init( char *, char *, BOOL );	// 
	virtual void Reset();						// ꥻå
	
	// ľܥؿ
	BYTE ReadMainRam( WORD addr ){ return MainRam[addr&(useextram ? 0x7fff : 0x3fff)]; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ǥХåѴؿ
	char *GetReadMemBlk( int, char * );
	char *GetWriteMemBlk( int, char * );
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	// ǥХ
	// OUTǥХID
	enum IDOut{ out92H=0, out93H };
	// INǥХID
	enum IDIn{};
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ɤǤSAVE
	BOOL DokoLoad( cIni * );	// ɤǤLOAD
	// ------------------------------------------
};


class cP6mem62 : public cP6memory, public Device {
protected:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IMAINRAM;
	
	BYTE *CGRom2;		// CG ROM2		60: -KB 62,66: 8KB
	BYTE *KanjiRom;		//  ROM		60: -KB 62,66:32KB
	BYTE *VoiceRom;		//  ROM	60: -KB 62,66:16KB
	
	BOOL cgrom;			// CG ROM  TRUE:N60⡼ FALSE:N60m⡼
	BOOL kj_rom;		// ROM,ROM  TRUE:ROM FALSE:ROM
	BOOL kj_LR;			// ROM  TRUE: FALSE:
	
	BYTE Rf0;			// ꥳȥݡ
	BYTE Rf1;			// ꥳȥݡ
	BYTE Rf2;			// ꥳȥݡ
	
	void SetMB_R( BYTE, BYTE );		// ꡼ɻΥ֥å(62,66Τ)
	void SetMB_W( BYTE );			// 饤ȻΥ֥å(62,66Τ)
	
	void SelectCGrom( int );		// CG ROM (62,66Τ)
	void SelectKanjiRom( BYTE );	// ROM  ROM ڤؤ(62,66Τ)
	
	// 8255ϴϢؿ
	void SetCGBank( BOOL );			// CG ROM BANK ڤؤ
	
	// ǥХ
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/Oؿ
	inline void Out92H( DWORD, BYTE data ){ SetCGBank( data&4 ? FALSE : TRUE ); }
	inline void Out93H( DWORD, BYTE data ){ if( (data&0x8e) == 4 ) SetCGBank( data&1 ? FALSE : TRUE ); }
	inline void OutC1H( DWORD, BYTE data ){ SelectCGrom( (data>>1)&1 ); }
	inline void OutC2H( DWORD, BYTE data ){ SelectKanjiRom( data ); }
	inline void OutC3H( DWORD, BYTE data ){}
	inline void OutF0H( DWORD, BYTE data ){ SetMB_R( data, Rf1 ); }
	inline void OutF1H( DWORD, BYTE data ){ SetMB_R( Rf0, data ); }
	inline void OutF2H( DWORD, BYTE data ){ SetMB_W( data ); }
	inline BYTE InF0H( DWORD ){ return Rf0; }
	inline BYTE InF1H( DWORD ){ return Rf1; }
	inline BYTE InF2H( DWORD ){ return Rf2; }
	
public:
	cP6mem62( const ID& );						// 󥹥ȥ饯
	virtual ~cP6mem62();						// ǥȥ饯
	
	virtual BOOL Init( char *, char *, BOOL );	// 
	void Reset();								// ꥻå
	
	// ľܥؿ
	BYTE ReadMainRom( WORD addr ) { return MainRom[addr&0x7fff]; }
	BYTE ReadMainRam( WORD addr ) { return MainRam[addr&0xffff]; }
	BYTE ReadCGrom2( WORD addr )  { return CGRom2[addr&0x1fff]; }
	BYTE ReadKanjiRom( WORD addr ){ return KanjiRom[addr&0x7fff]; }
	BYTE ReadVoiceRom( WORD addr ){ return VoiceRom[addr&0x3fff]; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ǥХåѴؿ
	char *GetReadMemBlk( int, char * );
	char *GetWriteMemBlk( int, char * );
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	// ǥХ
	// OUTǥХID
	enum IDOut{ out92H=0, out93H, outC1H, outC2H, outC3H, outF0H, outF1H, outF2H };
	// INǥХID
	enum IDIn{ inF0H=0, inF1H, inF2H };
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ɤǤSAVE
	BOOL DokoLoad( cIni * );	// ɤǤLOAD
	// ------------------------------------------
};


class cP6mem66 : public cP6mem62 {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IMAINRAM;
	
	void SelectKanjiRom( BYTE );		// ROM  ROM ڤؤ(62,66Τ)
	
public:
	cP6mem66( const ID& );				// 󥹥ȥ饯
	~cP6mem66();						// ǥȥ饯
	
	BOOL Init( char *, char *, BOOL );	// 
};



#endif	// MEMORY_H_INCLUDED
